#!/usr/bin/perl
##
## fetch_stat.pl
##
## Fetches and summarizes stats from non-snmp/web-smart switches
## Written by Joakim Andersson, joakim at iqcc dot se
##
## 1) Install the Perl-modules you might be missing
## 2) Adjust the settings in the my_switch/settings.txt
## 3) Run the script as perl fetch_stat.pl my_switch
##
## Arguments: 
## <dir>         - If first argument, read settings from other dir than my_switch/
##                 Allows polling of several switches 
## clear_stats	 - Clear the portstats when script starts
## no_retention  - Don't read retention at startup
## settings=file - Read settings from this file
##
## LEGACY:
## clear_session - Forces a cached session to be ignored, useful for testing only 
## clear_wraps   - Forgets about previous counter-wraps
##
## For changelog please visit http://www.iqcc.se
##

use strict;
use Time::HiRes;

our $progver  = '2.02b (2009-04-26)';
our $progname = 'fetch_stat.pl';

require 'lib/stdfunc.pm';

####################################################
# Setting up shared variables

our %TIME = ();
&upd_time();	# Setting time

our $SET = { 'VERBOSE' => 3 };	# Settings
our %STATS = ();			# To maintain delta-stats 
our %OLD_STATS = ();		# To cache previous values
our %PORTS = ();			# To maintain non-statistics info about each port
our %WRAPS = ();			# To maintain counter-wraps
our $SYSTEM = {};			# System-info

our %do_ports = {};			# We currently poll these ports
our @do_ports = ();			# Caches the portorder

#####################################################

$SYSTEM->{'BASEDIR'} = 'my_switch/';
$SYSTEM->{'SETTINGS_FILE'} = 'settings.txt';

# Allow first argument to be the directory
if ($ARGV[0] && $ARGV[0] !~ /\=/ && -d $ARGV[0])
{
	$SYSTEM->{'BASEDIR'} = $ARGV[0];
}

# Looking for valid args
my $clear_session = 0;
my $clear_stat = 0;
my $clear_wraps = 0;
my $no_retention = 0;

foreach (@ARGV)
{
	($_, my $value) = split(/\=/, $_, 2);
		
	if (/^\-*clear_stats?$/)
	{
		$clear_stat = 1;
	}
	elsif (/^\-*clear_sessions?$/)
	{
		$clear_session = 1;
	}
	elsif (/^\-*clear_wraps?$/)
	{
		$clear_wraps = 1;
	}
	elsif (/^\-*no_retention?$/)
	{
		$no_retention = 1;
	}
	elsif (/^\-*settings$/ && $value)
	{
		$SYSTEM->{'SETTINGS_FILE'} = $value;
	}
	elsif (/^\-*basedir$/ && $value)
	{
		$SYSTEM->{'BASEDIR'} = $value;
	}
}

$SYSTEM->{'BASEDIR'} .= '/' unless ($SYSTEM->{'BASEDIR'} =~ /\/$/);
$SYSTEM->{'SETTINGS_FILE'} = $SYSTEM->{'BASEDIR'} . $SYSTEM->{'SETTINGS_FILE'};

# Reading settings
# Will be refreshed when changed later
if (&refresh_settings() == 2)
{
	##
	## Errors in the settings!
	##
	
	die "Errors exists in '$SYSTEM->{'SETTINGS_FILE'}', unable to start $progname $progver!";
}

# Settings was OK
&log("$progname $progver started at $TIME{'date_time'}", $SET->{'VERBOSE'});

# Load the switch-specific module
$SET->{'MODULE'} .= '.pm' unless ($SET->{'MODULE'} =~ /\.pm$/);
foreach ("lib/$SET->{'MODULE'}", $SET->{'MODULE'},   'NA')
{
	if ($_ eq 'NA')
	{
		&log_n_die("Unable to find module $SET->{'MODULE'}! Please check your settings");
	}
	next unless (-e $_);
	
	# Located, loading
	#print "Loading switch-specific module $_...\n";
	require $_;
	last;
}

if ($SET->{'RETENTION_FILE'})
{
	##
	## Restoring retention
	##
	
	if ($no_retention)
	{
		print "Ignoring existing retention\n" if ($SET->{'VERBOSE'});
	}
	else
	{
		&load_retention();
	}
}

# LEGACY
# Reads cached cookie, unless already loaded from retention
if ($SET->{'COOKIE_FILE'} && !$SYSTEM->{'cookie'})
{
	open (COOKIE, $SET->{'COOKIE_FILE'}) || warn "Unable to open cookie-file $SET->{'COOKIE_FILE'}, first run ever?\n";
	$SYSTEM->{'cookie'} = <COOKIE>;
	close COOKIE;
}

if ($clear_session && $SYSTEM->{'cookie'})
{
	$SYSTEM->{'cookie'} = '';
	print "Cleared old session-cookie\n" if ($SET->{'VERBOSE'});
}
elsif ($SYSTEM->{'cookie'})
{
	$SYSTEM->{'cookie'} =~ tr/\n\r\t//d;
	print "Trying to reuse old session-cookie $SYSTEM->{'cookie'}\n" if ($SET->{'VERBOSE'});
	
	# Fetches system info
	&get_system_info();
}

##
## If we don't have a valid session login and fetch system info again
##

unless ($SYSTEM->{'cookie'})
{
	&login();
	&get_system_info();
}

&log("Found a $SYSTEM->{'model'} with $SYSTEM->{'numberOfPorts'} ports, firmware $SYSTEM->{'firmware'}", $SET->{'VERBOSE'});

##
## Clears stats of these ports if program was called with arg clear_stat
##

if ($clear_stat)
{
	print "Clearing stats for all ports...\n";
	&switch_clear_stat();
	
	%STATS = ();
	%OLD_STATS = ();
	%WRAPS = ();
}

##
## LEGACY
## Reading or clearing wraps
## Unless we already loaded if from retention
##

if ($SET->{'WRAPS_FILE'} && !%WRAPS)
{
	if ($clear_wraps || $clear_stat)
	{
		##
		## Removes cached wraps
		##
		
		if (-e $SET->{'WRAPS_FILE'})
		{
			&log("Clearing cached wraps", $SET->{'VERBOSE'});
			unlink $SET->{'WRAPS_FILE'};
		}
		
		%WRAPS = ();
	}
	else
	{
		##
		## Reads cached wraps
		##
		
		open (WRAPS, $SET->{'WRAPS_FILE'});
		while (<WRAPS>)
		{
			my ($name, $value) = split(/\t/, $_, 2);
			chop $value;
			
			$WRAPS{$name} = $value;
			#print "$name = $value\n";
		}
		close WRAPS;
	}
}

##
## Loops forever, hopefully
##

my @do_delta = qw(goodOctetRCV rxGood goodOctetSND txGood rx64Octets rx65TO127Octets rx128TO255Octets rx256TO511Octets rx512TO1023Octets rx1024ToMa );

our $loop = 0;
our $looptime = 'n/a';
my $got_delta = 0;

print "\n" if ($SET->{'VERBOSE'});	#Init complete\n
$| = 1;

while (1)
{
	$loop++;
	
	##
	## Refreshes settings if needed, but not on first loop
	##
	
	my $settings_changed = 0;
	$settings_changed = &refresh_settings() unless ($loop == 1);
	
	my $got_system_info = 0;
	$SYSTEM->{'real_refresh_rate'} = 0;
	
	if ($settings_changed == 2)
	{
		##
		## Errors!
		##
		
		&log("Errors found in the settings, keeping the old settings!", $SET->{'VERBOSE'});
		$settings_changed = 0;
	}
	elsif ($settings_changed)
	{
		##
		## New settings read, fetch portinfo to be able to
		## autoadd ports before we poll
		##
		
		&get_system_info();
		$got_system_info = 1;
	}
	
	##
	## Login if we miss a valid session
	##
	
	&login() unless ($SYSTEM->{'cookie'});
	
	# Remembers the time when this loop started
	my $start_time = [ Time::HiRes::gettimeofday() ];
	
	# Updates time
	&upd_time();
	
	##
	## Polling every defined port
	##
	
	my $wraps_occured = 0;
	my $is_first_poll = 1;
	print "Polling stats ($TIME{'time'})...\n" if ($SET->{'VERBOSE'} >= 2);
	
	foreach my $portnum (@do_ports)
	{
		##
		## Initiates the hashes
		##
		
		my $STAT = { 'time' => [ Time::HiRes::gettimeofday() ] };
		my $OLD_STAT = $STATS{$portnum};
		my $PORT = $PORTS{$portnum} ||= {};
		
		if ($SET->{'VERBOSE'} >= 2)
		{
			# Builds the info
			my $name = "$portnum:";
			print $name, ' ' x (6 - length($name));
			
			$name = $do_ports{$portnum};	#"$do_ports{$portnum}...";
			print $name, ' ' x ($SYSTEM->{'maxlen_portname'} - length($name) + 2);
		}
		
		##
		## Is it a switchport or a local interface?
		##
		
		if ($portnum =~ /^T?\d+$/)
		{
			##
			## Polls this port on the switch
			## Returns nada if we werent logged in
			##
			
			my $vars = &switch_poll_port($portnum, $is_first_poll);
			last unless ($vars);
			
			##
			## Handling wraps
			## Trying to handle clearing of the statstics in the switch
			##
			
			my %wrapped = ();
			
			if ($OLD_STAT)
			{
				my $wraps = 0;
				
				foreach my $name (sort keys %$vars)
				{
					my $value = $vars->{$name};
					my $old_value = $OLD_STAT->{"real:$name"};
					
					# Only check numerical values, and make
					# sure we got a old value
					next unless ($old_value =~ /^\d+$/);
					
					# So, is it smaller now?
					next unless ($value < $old_value);
					
					##
					## We have detected a (hopefully valid) wrap
					## With 32bit counters 4294967296 is the max value
					## If the switch uses 64bit counters then we will
					## probably not come here, and the values will
					## be too big to handle anyway
					##
					
					$wrapped{$name} = 1;
					
					next unless (++$wraps >= 3);
					
					##
					## Looks like a reset of statistics has been done
					## Or a reboot
					## Resetting our wrap-counter and old stats too
					##
					
					&log("A reset of the stats seems to have occured. Resetting our old stats and wraps too", $SET->{'VERBOSE'} >= 3);
					
					# Deleteing the old useless stat
					undef $OLD_STAT;
					delete $OLD_STATS{$portnum};
					
					# Cleaning old wraps for this port
					foreach (keys %WRAPS)
					{
						delete $WRAPS{$_} if (/^$portnum:/);
					}
					
					# Clearing logged wraps
					%wrapped = ();
					last;
				}
			}
			
			foreach my $name (sort keys %$vars)
			{
				# The real value without wraps
				# Also allows other programs to do access it via
				# external programs by loading the retention
				my $value = $STAT->{"real:$name"} = $vars->{$name};
				
				##
				## Adding previous valid wraps of this counter
				##
				
				if ($WRAPS{"$portnum:$name"})
				{
					$value += 4294967296 * $WRAPS{"$portnum:$name"};
				}
				
				# Did this counter wrap now?
				if ($wrapped{$name})
				{
					# Yes, increase wraps, add value
					my $tmp = ++$WRAPS{"$portnum:$name"};
					$value += 4294967296;
					$wraps_occured++;
					my $old_real = $OLD_STAT->{"real:$name"};
					
					&log("$do_ports{$portnum} (port $portnum): Wrap of $name has occured. $old_real -> $vars->{$name}. (tot $OLD_STAT->{$name} -> $value, wraps $tmp)", $SET->{'VERBOSE'} >= 3);
				}
				
				# Memorizing this counter
				$STAT->{$name} = $value;
			}
			
			##
			## Verify that we got valid input
			##
			
			unless (defined $STAT->{'goodOctetRCV'})
			{
				##
				## Hmmmmm, didn't find any of those expected variables
				##
				
				if ($portnum =~ /^T(\d+)$/ && ($SYSTEM->{'numberOfTrunks'} < $1))
				{
					# Port out of bounds
					&log_n_die("Unable to poll trunk $1 ($portnum), because max supported by switch is $SYSTEM->{'numberOfTrunks'}");
				}
				elsif ($portnum =~ /^(\d+)$/ && ($SYSTEM->{'numberOfPorts'} < $1))
				{
					# Trunk out of bounds
					&log_n_die("Unable to poll port $1, because max supported by switch is $SYSTEM->{'numberOfPorts'}");
				}
				
				# Should always recieve portIndex, just a failsafe
				&log("Bug? Unable to poll port $portnum, unknown cause", $SET->{'VERBOSE'});
				exit;
			}
		}
		elsif ($portnum =~ /^C\d+$/)
		{
			##
			## This was a local iptables' chain
			## Variables are added to $STAT if successful
			##
			
			my $def = ${$SET->{'CHAINS'}}{$portnum};
			unless (&poll_chain($def, $STAT))
			{
				##
				## The poll failed
				##
				
				&log("Poll of chain $portnum definition '$def' failed!");
				next;
			}
		}
		else
		{
			##
			## This was assumes to be a local interface
			## Variables are added to $STAT if successful
			##
			
			unless (&poll_interface($portnum, $STAT))
			{
				##
				## The poll failed
				##
				
				&log("Poll of local interface '$portnum' failed!");
				next;
			}
		}
		
		##
		## Do we have old info, so we can compute delta-values?
		##
		
		if (defined $OLD_STAT)
		{
			##
			## Computing delta values of some variables
			##
			
			# Elapsed time since last run
			$STAT->{'delta_time'} = Time::HiRes::tv_interval($OLD_STAT->{'time'}, $STAT->{'time'});
			
			# Real time since last loop
			$SYSTEM->{'real_refresh_rate'} = $STAT->{'delta_time'} if ($STAT->{'delta_time'} > $SYSTEM->{'real_refresh_rate'});
			
			if ($STAT->{'delta_time'} > 0)
			{
				##
				## Calcs delta for some variables
				##
				
				foreach (@do_delta)
				{
					my $diff = $STAT->{$_} - $OLD_STAT->{$_};
					$STAT->{"delta_$_"} = $diff / $STAT->{'delta_time'};
					
					if ($diff < 0)
					{
						##
						## Should never get a negative value!
						##
						
						&log("Negative delta! Port $portnum, variable $_: now '$STAT->{$_}', prev '$OLD_STAT->{$_}' = diff '$diff'", $SET->{'VERBOSE'});
					}
				}
				
				##
				## Calcs errors/sec per in/out
				##
				
				# error-pps in
				my $value = 0;
				$value += ($STAT->{$_} - $OLD_STAT->{$_}) foreach (qw(rxFCS rxUnderSize rxOverSize rxFragment rxJabber));
				$STAT->{'err_in_pps'} = $value / $STAT->{'delta_time'};
				
				# error-pps out
				$value = 0;
				$value += ($STAT->{$_} - $OLD_STAT->{$_}) foreach (qw(txDrop txCollisions));
				$STAT->{'err_out_pps'} = $value / $STAT->{'delta_time'};
				
				##
				## Show stats in the shell?
				##
				
				if ($SET->{'VERBOSE'} >= 2)
				{
					my @values = ();
					
					foreach (qw(goodOctetRCV rxGood goodOctetSND txGood))
					{
						my $value = $STAT->{"delta_$_"};
						
						if ($PORT->{'has_link'} || $value > 0)
						{
							if (/Octet/)
							{
								# Bytes -> bits
								$value = &shorten_num($value*8,undef,1024) . 'bps';
							}
							else
							{
								# Packets
								$value = &shorten_num($value,undef,1000) . 'pps';
							}
						}
						else
						{
							$value = '-';
						}
						push (@values, $value);
					}
					
					print "out:", ' ' x (8 - length($values[0])), $values[0];
					print " / ", ' ' x (8 - length($values[1])), $values[1];
					
					print "     in:", ' ' x (8 - length($values[2])), $values[2];
					print " / ", ' ' x (8 - length($values[3])), $values[3];
					
					my $tot_err_pps = $STAT->{'err_in_pps'}+$STAT->{'err_out_pps'};
					if ($tot_err_pps > 0)
					{
						$tot_err_pps = &shorten_num($tot_err_pps,undef,1000) . 'pps';
					}
					else
					{
						$tot_err_pps = '-';
					}
					print "     err:", ' ' x (8 - length($tot_err_pps)), $tot_err_pps;
					
					my $show = &shorten_num($STAT->{'delta_time'}, 6);
					print "     delta: ${show}s";
				}
			}
			
			##
			## Shall we write to RRD-file
			##
			
			if ($SET->{'RRD_FILE_PREFIX'} && !$SET->{"RRD_SKIP_PORT_$portnum"})
			{
				##
				## Yep, we shall save RRD data for this port
				##
				
				my $filename = $SET->{'RRD_FILE_PREFIX'} . "_$portnum.rrd";
				my $status = 1;
				
				unless (-e $filename)
				{
					# Doesn't exist, create it
					($status, my $message) = &create_rrd($filename);
					&log("Port $portnum: $message", $SET->{'VERBOSE'} >= 4);
					
					if ($status == 0)
					{
						# Will not try create RRD for this port again
						$SET->{"RRD_SKIP_PORT_$portnum"} = 1;
						&log("Create failed, disabling RRD for port $portnum", 1);
					}
				}
				
				if ($status == 1)
				{
					##
					## No errors found, writing
					##
					## Order of data:
					## kbps out and in
					## pps out and in
					## error-pps out and in
					##
					
					my @values = ();
					
					foreach (qw(delta_goodOctetSND delta_goodOctetRCV
								delta_txGood delta_rxGood
								err_out_pps err_in_pps
							))
					{
						my $value = $STAT->{$_};
						$value *= 8/1024 if (/Octet/);	# Octets -> kbit
						$value =~ s/(\.\d\d)\d+$/$1/;
						push(@values, $value);
					}
					
					# Uses the timestamp from our previous Time::HiRes gettimeofday
					# HOWEVER if we're syncing to RRD and we're in sync the loop-start-time is used
					my $timestamp = ${$STAT->{'time'}}[0];
					$timestamp = $TIME{'unixtime'} if ($SET->{'rrd_in_sync'});
					
					my $string = join(':', $timestamp, @values);
					
					##
					## Updates the RRD
					##
					
					RRDs::update($filename, $string);
					
					my $error = &RRDs::error;
					&log("Unable to update RRD '$filename' with '$string'!: $error", $SET->{'VERBOSE'} >= 1) if ($error);
				}
			}
			
			##
			## Is this port considered active?
			## Used when we hide inactive ports in the HTML
			##
			
			$PORT->{'is_active'} = $PORT->{'has_link'};
			if (!$PORT->{'is_active'} && ($STAT->{'delta_txGood'} || $STAT->{'delta_rxGood'}))
			{
				##
				## According to last polling of linkinfo the port is down, but
				## the count of packets travelling the port has increased since
				## last loop, so we assume it's gone up
				##
				
				$PORT->{'is_active'} = 1;
				$STAT->{'link_changed'} = 1;
			}
		}
				
		# Memorizes the previous stat for this port
		$OLD_STATS{$portnum} = $STATS{$portnum} if ($STATS{$portnum});
		$STATS{$portnum} = $STAT;
		$is_first_poll = 0;
		
		print "\n" if ($SET->{'VERBOSE'} >= 2);
	}
	
	##
	## Writes cached wraps
	##
	
	if ($wraps_occured > 0 && $SET->{'WRAPS_FILE'})
	{
		#&log("Saving wraps-cache", $SET->{'VERBOSE'} >= 2);
		
		open (WRAPS, ">$SET->{'WRAPS_FILE'}") || &log("Failed to save wraps-cache to file '$SET->{'WRAPS_FILE'}'!", $SET->{'VERBOSE'});
		foreach (sort keys %WRAPS)
		{
			print WRAPS "$_\t$WRAPS{$_}\n";
		}
		close WRAPS;
	}
	
	# Ends loop here if we don't have a valid session
	next unless ($SYSTEM->{'cookie'});
	
	##
	## Checking and logging significant info
	##
	
	foreach my $portnum (@do_ports)
	{
		my $STAT = $STATS{$portnum};
		next unless ($STAT->{'delta_time'});
		
		my $OLD_STAT = $OLD_STATS{$portnum};
		my $PORT = $PORTS{$portnum};
		
		##
		## Calculating pps for the less important counters
		##
		
		foreach (qw(txBroadcast txMulticast rxBroadcast rxMulticast))
		{
			my $rate = &shorten_num(($STAT->{$_} - $OLD_STAT->{$_}) / $STAT->{'delta_time'}, 3, 1000);
			
			# Info for ALT in the HTML
			$PORT->{"$_:INFO"} = "Rate ${rate}pps";
		}
		
		##
		## Checking and logging increased errors
		##
		
		foreach (qw(txDrop txCollisions rxFCS rxUnderSize rxOverSize rxFragment rxJabber))
		{
			my $diff = $STAT->{$_} - $OLD_STAT->{$_};
			next unless ($diff > 0);
			
			my $deltatime = &shorten_num($STAT->{'delta_time'}, 4);	# Only stripping decimals
			my $rate = &shorten_num($diff / $STAT->{'delta_time'}, 3, 1000);
			
			# Log and possibly print if verbose >= 3
			&log("$do_ports{$portnum} (port $portnum): $_ has increased by $diff the last $deltatime secs (rate ${rate}pps)", $SET->{'VERBOSE'} >= 3);
			
			# Info for ALT in the HTML
			$PORT->{"$_:INFO"} = "Last at $TIME{'date_time'}, rate was ${rate}pps";
		}
	}
	
	##
	## We doesn't fetch system and linkinfo every loop, since it takes a while
	##
	
	unless ($got_system_info)
	{
		if ($loop % $SET->{'POLL_LINKINFO_EVERY_NTH_LOOP'} == 0)
		{
			##
			## Also fetches general system-info now
			##
			
			&get_system_info();
		}
		else
		{
			# Won't fetch, so we assumes the switch has been up
			# We assumes that the whole loop will take refresh_rate seconds 
			$SYSTEM->{'uptime'} += $SET->{'REFRESH_RATE'};
		}
	}
	
	# Ends loop here if we don't have a valid session
	next unless ($SYSTEM->{'cookie'});
	
	##
	## We doesn't fetch mac-address-table every loop, since it takes a while
	##
	
	if ($SET->{'POLL_MACS_EVERY_NTH_LOOP'} && (($loop-1) % $SET->{'POLL_MACS_EVERY_NTH_LOOP'} == 0))
	{
		##
		## Fetches mac-address-table
		##
		
		&poll_macs();
	}
	
	# Ends loop here if we don't have a valid session
	next unless ($SYSTEM->{'cookie'});

	##
	## This loop seems to be done and hopefully successful
	## Logging out, if wanted
	##
	
	&logout() if ($SET->{'ALWAYS_LOGOUT'});
	
	##
	## Save retention if wanted
	## Saves after logout so we don't have a invalid cookie in retention
	##
	
	if ($SET->{'RETENTION_FILE'} && (($loop+1) % $SET->{'RETENTION_EVERY_NTH_LOOP'} == 0))
	{
		##
		## Saves retention
		##
		
		&save_retention();
	}
	
	##
	## Generates graphs
	##
	
	if ($SET->{'HTML_FILE'} && $SET->{'GRAPH_BPS_EVERY_NTH_LOOP'} && (($loop+1) % $SET->{'GRAPH_BPS_EVERY_NTH_LOOP'} == 0))
	{
		&generate_rrd_graphs('bps');
	}
	
	if ($SET->{'HTML_FILE'} && $SET->{'GRAPH_PPS_EVERY_NTH_LOOP'} && (($loop+1) % $SET->{'GRAPH_PPS_EVERY_NTH_LOOP'} == 0))
	{
		&generate_rrd_graphs('pps');
	}
	
	##
	## Generates web-page if wanted
	##
	
	if ($SET->{'HTML_FILE'})
	{
		$looptime = Time::HiRes::tv_interval($start_time, [ Time::HiRes::gettimeofday() ]);
		&generate_html();
	}
	
	##
	## If single loop we break now
	## Retention will keep track of elapsed time
	##
	
	if ($SET->{'SINGLE_LOOP'} && $SYSTEM->{'real_refresh_rate'})
	{
		print "Done for now, SINGLE_LOOP is in effect. See you back soon!\n";
		last;
	}
	
	##
	## Pauses
	##
	
	$looptime = Time::HiRes::tv_interval($start_time, [ Time::HiRes::gettimeofday() ]);
	my $sleep = $SET->{'REFRESH_RATE'} - $looptime;
	
	##
	## RRD creates "sharper" graphs if data is polled near each RRD_STEP-interval
	## If wanted we try to 'slide' to stay near that interval
	##
	
	if ($SET->{'RRD_FILE_PREFIX'} && $SET->{'SYNC_SLEEP_TO_RRD'})
	{
		# If using the default RRD_STEP 30s, we get a value between 0 and +30
		# 0 is the ideal point
		# Negative means we need to sleep a bit shorter to get near 0
		# Positive means we need to sleep a bit longer to get near 0
		my $half_step = $SET->{'RRD_STEP'} / 2;
		my $diff = $half_step - ($TIME{'unixtime'}+$half_step) % $SET->{'RRD_STEP'};
		
		if ($diff != 0)
		{
			##
			## Need adjusting
			##
			
			# Max 5 seconds each loop
			$diff = 5 if ($diff > 5);
			$diff = -5 if ($diff < -5);
			$sleep += $diff;
			
			$diff = '+'.$diff if ($diff > 0);
			&log("Adjusting sleep with $diff secs to sync with RRD_STEP = $SET->{'RRD_STEP'} secs", $SET->{'VERBOSE'});
			$SET->{'rrd_in_sync'} = 0;
		}
		elsif ($SET->{'rrd_in_sync'} != 1)
		{
			print "Sync with RRD_STEP = $SET->{'RRD_STEP'} secs is OK\n" if ($SET->{'VERBOSE'} >= 2);
			$SET->{'rrd_in_sync'} = 1;
		}
	}
	
	# Always sleeps atleast 2 secs!
	$sleep = 2 if ($sleep < 1);
	
	print "Loop $loop took $looptime secs, sleeps $sleep secs...\n"  if ($SET->{'VERBOSE'});
	print "\n" if ($SET->{'VERBOSE'} > 1);
	Time::HiRes::sleep($sleep);
}

exit;

