#!/usr/bin/perl
##
## Common functions used by fetch_stat.pl
##

use strict;
use Time::HiRes;
use Storable qw(lock_store lock_retrieve);

##############################################

our $progver;		# Main Version
our $progname;		# Main Name
our $modver;		# Module Version
our $modname;		# Module Name
our $modcreator;	# Module Creator

our %TIME;			# Time
our $SET;			# Settings
our %STATS;			# To maintain delta-stats 
our %OLD_STATS;		# To cache previous values
our %PORTS;			# To maintain non-statistics info about each port
our %WRAPS;			# To maintain counter-wraps
our $SYSTEM;		# System-info

our %do_ports = {};
our @do_ports = ();

our $loop;
our $looptime;

my $TR_COLOR = '#FFFF80';
my %graph_names = ('bps' => 'bits per second', 'pps' => 'packets per second');
my %graph_rev_type = ('bps' => 'pps', 'pps' => 'bps');

##
## Update our global %TIME variable 
##

sub upd_time
{
	my $unixtime = shift || time;

	my ($sec,$min,$hour,$daynum,$monnum,$year,$weekdaynum) = localtime($unixtime);
	$sec = "0$sec" if ($sec < 10);
	$min = "0$min" if ($min < 10);
	$hour = "0$hour" if ($hour < 10);
	my $day = $daynum;
	$day = "0$day" if ($day < 10);
	my $mon = $monnum + 1;
	$mon = "0$mon" if ($mon < 10);
	$year += 1900;

	%TIME = (sec => $sec, min => $min, hour => $hour, day => $day, daynum => $daynum, weekdaynum => $weekdaynum, mon => $mon, monnum => $monnum, year => $year, unixtime => $unixtime, date_time => "$year-$mon-$day $hour:$min:$sec", date => "$year-$mon-$day", 'time' => "$hour:$min:$sec");
}

##
## Adds spaces to the numbers for clarity
## 12345678 -> 12'345'678
##

sub add_sep
{
	my $value = shift;
	$value =~ s/(\.\d+)$//;
	my $decimals = ($1 && $value < 10) ? $1 : '';
	$value = reverse $value;
	
	$value =~ s/(\d\d\d)/$1\'/g;
	$value =~ s/\'$//;
	return (reverse $value) . substr($decimals, 0, 3);
}

##
## Rewrites 1024 as 1k, with a given number of digits
##

sub shorten_num
{
	my $num = shift;
	my $digits = shift || 3;
	my $base = shift || 1024;
	my $unit = '';
	
	foreach (qw(k M G T P E Z Y))
	{
		last if ($num < $base);
		
		$unit = $_;
		$num /= $base;
	}
	
	$num = ($num >= 1000) ? substr($num, 0, 4) : substr($num, 0, $digits);
	$num =~ s/(\.\d*)0+$/$1/;
	$num =~ s/\.$//;
	
	return $num . $unit;
}

##
## Takes seconds and make them readable
##

sub pretty_time
{
	my $secs = shift;
	
	my $string = '';
	my $days = int($secs / 86400);
	if ($days)
	{
		$secs -= $days*86400;
		$string .= "${days}d";
	}
	my $hours = int($secs / 3600);
	if ($hours || $days)
	{
		$secs -= $hours*3600;
		$string .= "${hours}h";
	}
	my $mins = int($secs / 60);
	$string .= "${mins}m";
	
	return $string;
}

##
## Poll local interface
## We could easily have done this buffered instead
##

sub poll_interface
{
	my $interface = shift;
	my $STAT = shift || {};
	my $ok = 0;
	
	if ($SET->{'USE_ETHTOOL'})
	{
		##
		## Poll using ethtool, hopefully
		##

		unless (open (ETHTOOL, "$SET->{'USE_ETHTOOL'} -S $interface |"))
		{
			##
			## Unable to poll successfully!
			##
			
			&log("Unable to run command '$SET->{'USE_ETHTOOL'} -S $interface'! Disabling polling using ethtool", 1);
			$SET->{'USE_ETHTOOL'} = '';
		}
		else
		{
			##
			## We were able to run the command (hopefully)
			##
			
			my %E = ();
			
			while (<ETHTOOL>)
			{
				#     rx_packets: 6207689
				#     tx_packets: 2174409
				#     rx_bytes: 4192981116
				
				chomp;
				s/^\s*//;
				my ($name, $value) = split(/\s*\:\s*/, $_, 2);
				next unless ($value =~ /^\d+$/);	# Makes sure we don't take crap
				
				$E{$name} = $value;
			}
			close ETHTOOL;
			
			##
			## Sanity-check
			##
			
			my $not_ok = 0;
			
			foreach (qw(rx_bytes rx_packets tx_bytes tx_packets))
			{
				next if (defined $E{$_});
				
				&log("Ethtool didn't provide expected variable $_!", 1);
				$not_ok++;
			}
			
			if ($not_ok)
			{
				&log("Disabling polling using ethtool, ($SET->{'USE_ETHTOOL'} -S $interface)", 1);
				$SET->{'USE_ETHTOOL'} = '';
			}
			else
			{
				##
				## Okay, I'm not certain about the more unusual errors like FCS
				##
				
				$STAT->{'goodOctetRCV'}	= $E{'rx_bytes'};
				$STAT->{'rxGood'}		= $E{'rx_packets'};
				$STAT->{'rxBroadcast'}	= $E{'rx_broadcast'} if (defined $E{'rx_broadcast'});
				$STAT->{'rxMulticast'}	= $E{'rx_multicast'} if (defined $E{'rx_multicast'});
				$STAT->{'rxFCS'}		= $E{'rx_align_errors'} + $E{'rx_crc_errors'} if (defined $E{'rx_align_errors'} || defined $E{'rx_crc_errors'});
				$STAT->{'rxUnderSize'}	= $E{'rx_short_length_errors'} if (defined $E{'rx_short_length_errors'});
				$STAT->{'rxOverSize'}	= $E{'rx_long_length_errors'} if (defined $E{'rx_long_length_errors'});
				$STAT->{'rxFragment'}	= '';	# Available?
				$STAT->{'rxJabber'}		= '';	# Available?
				
				$STAT->{'goodOctetSND'}	= $E{'tx_bytes'};
				$STAT->{'txGood'}		= $E{'tx_packets'};
				$STAT->{'txBroadcast'}	= $E{'tx_broadcast'} if (defined $E{'tx_broadcast'});
				$STAT->{'txMulticast'}	= $E{'tx_multicast'} if (defined $E{'tx_multicast'});
				$STAT->{'txDrop'}		= $E{'tx_dropped'} if (defined $E{'tx_dropped'});
				$STAT->{'txCollisions'}	= $E{'collisions'} if (defined $E{'collisions'});
				
				$ok = 1;
			}
		}
	}
	
	unless ($SET->{'USE_ETHTOOL'})
	{
		##
		## Don't poll using ethtool
		## Opens /proc/net/dev to read some statistics
		##
		
		unless (open (DEV, '/proc/net/dev'))
		{
			&log("Unable to open '/proc/net/dev'!", 1);
			return 0;
		}
		
		# Searching for this interface
		while (<DEV>)
		{
			#Inter-|   Receive                                                |  Transmit
			# face |bytes    packets errs drop fifo frame compressed multicast|bytes    packets errs drop fifo colls carrier compressed
			#  eth0:230864891  964901    0    0    0     0          0         0 867172607  936222    0    0    0     0       0          0
			
			if (/^\s*$interface\:\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)$/)
			{
				##
				## Remembering some data
				##
				
				$STAT->{'goodOctetRCV'}	= $1;
				$STAT->{'rxGood'}		= $2;
				$STAT->{'rxFCS'}		= $6;
				$STAT->{'rxUnderSize'}	= '';
				$STAT->{'rxOverSize'}	= '';
				$STAT->{'rxFragment'}	= '';
				$STAT->{'rxJabber'}		= '';
				
				$STAT->{'goodOctetSND'}	= $9;
				$STAT->{'txGood'}		= $10;
				$STAT->{'txDrop'}		= $12;
				$STAT->{'txCollisions'}	= $14;
				
				$ok = 1;
				
				last;
			}
		}
		close DEV;
	}
	
	return $ok;
}

##
## Polls values from a local iptables' chain
##

sub poll_chain
{
	my $chain = shift;
	my $STAT = shift || {};
	my $ok = 0;
	
	$STAT->{'goodOctetRCV'}	= 0;
	$STAT->{'rxGood'}		= 0;
	$STAT->{'rxFCS'}		= '';
	$STAT->{'rxUnderSize'}	= '';
	$STAT->{'rxOverSize'}	= '';
	$STAT->{'rxFragment'}	= '';
	$STAT->{'rxJabber'}		= '';
	
	$STAT->{'goodOctetSND'}	= 0;
	$STAT->{'txGood'}		= 0;
	$STAT->{'txDrop'}		= '';
	$STAT->{'txCollisions'}	= '';
	
	##
	## Reads info from iptables
	## chain1,1,chain2,1
	##
	
	my @def = split(/\s*,\s*/, $chain);
	my @chains = (['OUT',$def[0],$def[1]], ['IN',$def[2],$def[3]]);
	
	foreach my $ref (@chains)
	{
		my $type = $$ref[0];
		my $chain = $$ref[1];
		my $row = $$ref[2] || -1;	# Defaults to last row
		next unless ($chain);
		
		unless (open (CHAIN, "iptables -vxn -L $chain |"))
		{
			&log("Unable to execute command 'iptables -vxn -L $chain'!", 1);
			return 0;
		}
		
		my @lines = <CHAIN>;
		close CHAIN;
		
		# Make -1 become the last actual row
		# If 4 lines total, 4-1 = 3 = lines[3] = last row
		$row = scalar @lines + $row if ($row < 0);
		
		# Pick the line to parse
		my $line = $lines[$row];
		
		# Parse that line, expect the first two lines to be packets and bytes
		#     7477   478570 RETURN     all  --  *      *       0.0.0.0/0            0.0.0.0/0
		unless ($line =~ /^\s*(\d+)\s+(\d+)\s/)
		{
			&log("Parsing chain '$chain'; row $row contained '$line', unable to parse!", 1);
			return 0;
		}
		
		#print "$type: $chain - row $row = $line -> pkts $1, bytes $2\n";
		
		if ($type eq 'OUT')
		{
			$STAT->{'goodOctetSND'}	= $2;
			$STAT->{'txGood'}		= $1;
		}
		else
		{
			$STAT->{'goodOctetRCV'}	= $2;
			$STAT->{'rxGood'}		= $1;
		}
	}
	return 1;
}

##
## Poll macs
##

sub poll_macs
{
	print "Polling mac-address-table...\n" if ($SET->{'VERBOSE'} >= 2);
	my $port_macs = &switch_poll_macs();
	
	unless ($port_macs)
	{
		# Failed!
		&log("Polling mac-address-table failed, not supported by switch/module? Disabling", 1);
		$SET->{'POLL_MACS_EVERY_NTH_LOOP'} = 0;
	}
	else
	{
		my %arp = ();
		
		if ($SET->{'LOOKUP_MAC_IN_ARP_CMD'})
		{
			# arp -a should work in both windows and unix
			foreach (split(/[\n\r]+/, `$SET->{'LOOKUP_MAC_IN_ARP_CMD'}`))
			{
				# squeezebox-tvrum.local.iqcc.se (192.168.0.19) at 00:04:20:12:5E:0C [ether] on eth0
				if (/^\s*(.+)\s+\(([\d\.]+)\) at ([a-fA-F0-9\:]+)\s/)
				{
					my $name = ($1 ne '?') ? $1 : '';
					my $ip = $2;
					my $mac = uc $3;
					#next if ($mac eq '00:00:00:00:00:00');
					$arp{$mac} = "$name ($ip)" if ($name && $ip);
					$arp{$mac} = "$name" if ($name && !$ip);
					$arp{$mac} = "$ip" if (!$name && $ip);
				}
				#   192.168.0.3           00-15-17-14-95-44     dynamisk
				elsif (/^\s*(\d\.)\s+([a-fA-F0-9\-]+)\s/)
				{
					my $mac = uc $2;
					$mac =~ s/\-/\:/g;
					next if ($mac eq '00:00:00:00:00:00');
					$arp{$mac} = $1;
				}
			}
		}
		
		# Parsing our intermediate table and stores into PORT
		foreach my $portnum (keys %$port_macs)
		{
			# Info for the HTML
			my $PORT = $PORTS{$portnum} || next;
			my $macs = $port_macs->{$portnum};
			$PORT->{"macs"} = "";
			$PORT->{"macs_count"} = 0;
			#delete $PORT->{"mac_single_desc"};
			
			foreach my $mac (sort keys %$macs)
			{
				next if (++$PORT->{"macs_count"} > 10);
				
				my $note = $macs->{$mac};
				my $desc = $SET->{'MACS'}->{$mac} || '';
				$desc = "$desc ($note)" if ($desc && $note);
				$desc = $note if (!$desc && $note);
				
				my $real = '';
				$real = " - $arp{$mac}" if ($arp{$mac});
				$PORT->{"macs"} .= "$mac$real";
				$PORT->{"macs"} .= " = $desc" if ($desc);
				$PORT->{"macs"} .= "\n";
			}
			$PORT->{"macs"} .= '...' if ($PORT->{"macs_count"} > 10);
			$PORT->{"macs"} =~ s/\n$//;
		}
	}
}

##
## Creates a empty RRD-file
##

sub create_rrd
{
	my $filename = shift || &error("No filename given to create as a RRD-file!");
	
	# Will not overwrite a existing file
	return (1,"File '$filename' already existed!") if (-e $filename);
	
	my $heartbeat = $SET->{'RRD_STEP'}*10;
	
	# Creating RRD-file
	RRDs::create(
		$filename,
		'--step',
		$SET->{'RRD_STEP'},
		"DS:kbps_out:GAUGE:$heartbeat:0:1000000",
		"DS:kbps_in:GAUGE:$heartbeat:0:1000000",
		"DS:pps_out:GAUGE:$heartbeat:0:10000000",
		"DS:pps_in:GAUGE:$heartbeat:0:10000000",
		"DS:err_pps_out:GAUGE:$heartbeat:0:10000000",
		"DS:err_pps_in:GAUGE:$heartbeat:0:10000000",
		@{$SET->{'RRD_RRA'}}
	);
	
	my $error = &RRDs::error;
	return(0, "Unable to create RRD-file '$filename'!: $error") if ($error);
	
	my $size = -s $filename;
	unless ($size > 0)
	{
		unlink $filename;	# Removes the file so we won't try to write to it later
		return(0, "Unable to create RRD-file '$filename'!: Zero size file");
	}
	
	return (1, "File '$filename' created, size $size bytes");
}

##
## Generate graphs
##

sub generate_rrd_graphs
{
	my $type = shift || 'bps';	# bps or pps
	
	my $uctype = uc $type;
	print "Graphing $uctype...\n" if ($SET->{'VERBOSE'} >= 2);
	
	##
	## Yep, let's create a bunch of files 
	##
	
	$SET->{'HTML_FILE'} =~ /^(.*\/)?([^\/]+)(\.[^.]+)$/;
	my $path = $1;		# path/to/
	my $base = $2;		# stats-file
	my $suffix = $3;	# .html
	
	#print "$path - $base - $suffix\n";
	
	# How often the graph-page will change
	my $refresh = $SET->{'REFRESH_RATE'} * $SET->{'GRAPH_' . $uctype . '_EVERY_NTH_LOOP'};
	
	# Shall we also graph the other type? Otherwise no links
	my $graph_other_type = $SET->{'GRAPH_' . uc $graph_rev_type{$type} . '_EVERY_NTH_LOOP'};
	
	my $ucname = ucfirst $graph_names{$type};
	my $rev = $graph_rev_type{$type};	# bps -> pps, pps -> bps
	my $switch = $SYSTEM->{'sysname'} || $SET->{'SWITCH'};
	
	my @index_graphs = ();		# We will create the page last
	
	##
	## Creates a overview for this type
	##
	
	foreach my $portnum (@do_ports)
	{
		##
		## Does each port
		##
		
		# Skip this port
		next if ($SET->{"RRD_SKIP_PORT_$portnum"});
		
		my $rrd_filename = $SET->{'RRD_FILE_PREFIX'} . "_$portnum.rrd";
		next unless (-s $rrd_filename);	# Will not proceed unless the rrd-data exist
		
		##
		## Writes indexfile for each port that will contain the graphs
		##
		
		my @port_graphs = ();		# We will create the page last
		
		my $rrd_port_file = "${base}_${type}_${portnum}$suffix";
		$PORTS{$portnum}->{"rrd_${type}_file"} = $rrd_port_file;	# Remembering
		
		##
		## Creating graphs
		##
		
		my @graph_args = ('--interlaced','--lazy');
		push (@graph_args, '--slope-mode') if ($SET->{'GRAPH_SLOPE_MODE'});
		
		foreach my $graph_ref (@{$SET->{'GRAPHS'}})
		{
			my $i			= $graph_ref->[0];
			my $graph_name	= $graph_ref->[1];
			my $graph_secs	= $graph_ref->[2];
			
			my $image_filename = "${base}_${type}_$portnum-$i.png";
			#unlink $image_filename if ($erase_images);
			
			##
			## Mega graph command
			##
			
			my $averages = my $xsize = my $ysize = 0;
			
			if ($type eq 'bps')
			{
				##
				## Bits/s
				##
				
				($averages,$xsize,$ysize) = RRDs::graph("$path$image_filename",
					"--start=end-$graph_secs",
					"--title=$do_ports{$portnum} ($portnum) on $switch - $graph_name",
					"--vertical-label=$graph_names{$type}",
					'--base=1024',
					'--imgformat=PNG',
					"--width=$SET->{'GRAPH_WIDTH'}",
					"--height=$SET->{'GRAPH_HEIGHT'}",
					'--alt-y-mrtg',
					@graph_args,
					"DEF:b=$rrd_filename:kbps_out:AVERAGE",
					"DEF:a=$rrd_filename:kbps_in:AVERAGE",
					'CDEF:B=b,1024,*',
					'CDEF:A=a,1024,*',
					'AREA:B#00FF00:bit/s OUT',
					'VDEF:B_AVERAGE=B,AVERAGE',
					'GPRINT:B_AVERAGE:Avg\: %7.2lf%s ',
					'VDEF:B_MAX=B,MAXIMUM',
					'GPRINT:B_MAX:Max\: %7.2lf%s ',
					'VDEF:B_LAST=B,LAST',
					'GPRINT:B_LAST:Last\: %7.2lf%s\n',
					'LINE1:A#0000FF:bit/s IN ',
					'VDEF:A_AVERAGE=A,AVERAGE',
					'GPRINT:A_AVERAGE:Avg\: %7.2lf%s ',
					'VDEF:A_MAX=A,MAXIMUM',
					'GPRINT:A_MAX:Max\: %7.2lf%s ',
					'VDEF:A_LAST=A,LAST',
					'GPRINT:A_LAST:Last\: %7.2lf%s'
				);
			}
			else
			{
				##
				## Packets/s
				##
				
				($averages,$xsize,$ysize) = RRDs::graph("$path$image_filename",
					"--start=end-$graph_secs",
					"--title=$do_ports{$portnum} ($portnum) on $switch - $graph_name",
					"--vertical-label=$graph_names{$type}",
					'--base=1000',
					'--imgformat=PNG',
					"--width=$SET->{'GRAPH_WIDTH'}",
					"--height=$SET->{'GRAPH_HEIGHT'}",
					'--alt-y-mrtg',
					@graph_args,
					"DEF:d=$rrd_filename:pps_out:AVERAGE",
					"DEF:c=$rrd_filename:pps_in:AVERAGE",
					"DEF:f=$rrd_filename:err_pps_out:AVERAGE",
					"DEF:e=$rrd_filename:err_pps_in:AVERAGE",
					'AREA:d#00FF00:Pkts/s OUT    ',
					'VDEF:d_AVERAGE=d,AVERAGE',
					'GPRINT:d_AVERAGE:Avg\: %7.2lf%s',
					'VDEF:d_MAX=d,MAXIMUM',
					'GPRINT:d_MAX:Max\: %7.2lf%s',
					'VDEF:d_LAST=d,LAST',
					'GPRINT:d_LAST:Last\: %7.2lf%s\n',
					'LINE1:c#0000FF:Pkts/s IN     ',
					'VDEF:c_AVERAGE=c,AVERAGE',
					'GPRINT:c_AVERAGE:Avg\: %7.2lf%s',
					'VDEF:c_MAX=c,MAXIMUM',
					'GPRINT:c_MAX:Max\: %7.2lf%s',
					'VDEF:c_LAST=c,LAST',
					'GPRINT:c_LAST:Last\: %7.2lf%s\n',
					'LINE1:f#FFA500:Err pkts/s OUT',
					'VDEF:f_AVERAGE=f,AVERAGE',
					'GPRINT:f_AVERAGE:Avg\: %7.2lf%s',
					'VDEF:f_MAX=f,MAXIMUM',
					'GPRINT:f_MAX:Max\: %7.2lf%s',
					'VDEF:f_LAST=f,LAST',
					'GPRINT:f_LAST:Last\: %7.2lf%s\n',
					'LINE1:e#FF0000:Err pkts/s IN ',
					'VDEF:e_AVERAGE=e,AVERAGE',
					'GPRINT:e_AVERAGE:Avg\: %7.2lf%s',
					'VDEF:e_MAX=e,MAXIMUM',
					'GPRINT:e_MAX:Max\: %7.2lf%s',
					'VDEF:e_LAST=e,LAST',
					'GPRINT:e_LAST:Last\: %7.2lf%s'
				);
			}
			
			my $error = &RRDs::error;
			&log("Unable to create graph '$path$image_filename'! $error", 1) if ($error);
			
			##
			## We create graphs first and then the page
			##
			
			push (@port_graphs,  "<img src=\"$image_filename\" width=$xsize height=$ysize>");
			push (@index_graphs, "<a href=\"$rrd_port_file\"><img src=\"$image_filename\" width=$xsize height=$ysize border=0></a>") if ($i == $SET->{'GRAPH_OVERVIEW'});
			
			# This graph is done
		}
		
		##
		## All graphs for this port done, create the page
		##
		
		my $switch_link = ($graph_other_type) ? "<br><a href=\"${base}_${rev}_$portnum$suffix\">Switch to $graph_names{$rev}</a>" : '';

		open(HTML, ">$path$rrd_port_file");
		print HTML <<END;
<html>
<head>
<title>$ucname for $do_ports{$portnum} (port $portnum) on $SET->{'SWITCH'} - $SYSTEM->{'sysname'} $SYSTEM->{'location'}</title>
<meta http-equiv="Refresh" content="$refresh">
</head>

<body bgcolor="White" text="Black" link="Black" vlink="Black" alink="Red" style="font-family: Verdana, Geneva, Arial, Helvetica, sans-serif; font-size: 9;">
<center>
<div style="font-size: 13; font-weight: bold;">$ucname - $do_ports{$portnum} (port $portnum)</div>
$switch_link
<br><a href="${base}_$type$suffix">Show overview for $graph_names{$type}</a>

<p>
END
			
		my $i = 0;
		foreach (@port_graphs)
		{
			print HTML $_;
			print HTML "<br>\n" if ($SET->{'GRAPHS_PER_ROW'} && (++$i % $SET->{'GRAPHS_PER_ROW'} == 0));
		}
		
		print HTML <<END;
<p>Page updated $TIME{'date_time'}, refreshes every $refresh sec
		
</center>
</body>
</html>
END
		close HTML;
		# Page for BPS or PPS done fr this port
	}
	
	##
	## Writing the index-page for this type
	##
	
	my $type_index_file = "${base}_${type}$suffix";
	my $switch_link = ($graph_other_type) ? "<br><a href=\"${base}_$rev$suffix\">Switch to $graph_names{$rev}</a>" : '';
	
	open(HTML, ">$path$type_index_file");
	print HTML <<END;
<html>
<head>
<title>Overview $graph_names{$type} on $SET->{'SWITCH'} - $SYSTEM->{'sysname'} $SYSTEM->{'location'}</title>
<meta http-equiv="Refresh" content="$refresh">
</head>

<body bgcolor="White" text="Black" link="Black" vlink="Black" alink="Red" style="font-family: Verdana, Geneva, Arial, Helvetica, sans-serif; font-size: 9;">
<center>
<div style="font-size: 13; font-weight: bold;">$ucname - Overview</div>
$switch_link
<br><a href="$base$suffix">Return to statistics</a>
<p>
END
	
	my $i = 0;
	foreach (@index_graphs)
	{
		print HTML $_;
		print HTML "<br>\n" if ($SET->{'GRAPHS_PER_ROW'} && (++$i % $SET->{'GRAPHS_PER_ROW'} == 0));
	}
	
	print HTML <<END;
<p>Page updated $TIME{'date_time'}, refreshes every $refresh sec

</center>
</body>
</html>
END
	close HTML;
	
	##
	## All ports done
	##
	
	return 1;
}

##
## Reads the settings
## Returns 0 when fails
##

sub refresh_settings
{
	##
	## Checks if the settings-file has changed
	##
	
	if ($SYSTEM->{'SETTINGS_FILES'})
	{
		##
		## Summarize modified-at from the old files
		## Will also react to deleted files
		##
		
		my $file_modified = 0;
		
		# Adding modified-time for included settings too
		foreach my $file (@{$SYSTEM->{'SETTINGS_FILES'}})
		{
			$file_modified += (stat($file))[9];
			#print "CHECK $file - $file_modified\n";
		}
		
		if ($file_modified eq $SET->{'_modified'})
		{
			# 0 = No change, all files are the same
			#print "NO CHANGE, $file_modified == $SET->{'_modified'}\n";
			return 0;
		}
	}
	
	##
	## Changed, re-reading settings
	##
	
	my %NEW_SET = ();
	$NEW_SET{'_modified'} = 0;
	$SYSTEM->{'SETTINGS_FILES'} = [];
	
	&read_settings($SYSTEM->{'SETTINGS_FILE'}, \%NEW_SET) || &log_n_die("Unable to open settings-file '$SYSTEM->{'SETTINGS_FILE'}'!");
	
	# Prevents attempts to re-read settings if errors are found
	$SET->{'_modified'} = $NEW_SET{'_modified'};
	
	##
	## Checking basic required variables
	##
	
	my $errors = 0;
	
	foreach (qw(MODULE SWITCH USERNAME PASSWORD REFRESH_RATE POLL_LINKINFO_EVERY_NTH_LOOP
				VERBOSE GRAPHS_PER_ROW GRAPH_WIDTH GRAPH_HEIGHT))
	{
		next if (defined $NEW_SET{$_});
		
		&log("Error: Required setting $_ is not defined", 1);
		$errors++;
	}
	
	##
	## Checking numerical-only settings
	##
	
	foreach (qw(REFRESH_RATE POLL_LINKINFO_EVERY_NTH_LOOP VERBOSE
				AUTOADD_ACTIVE_PORTS HIDE_INACTIVE_PORTS
				GREY_BELOW_PPS BOLD_ABOVE_PPS GREY_BELOW_BPS
				BOLD_ABOVE_BPS GREY_BELOW_PROC
				GRAPH_BPS_EVERY_NTH_LOOP GRAPH_PPS_EVERY_NTH_LOOP
				RRD_STEP GRAPH_WIDTH GRAPH_HEIGHT GRAPHS_PER_ROW
				GRAPH_SLOPE_MODE POLL_MACS_EVERY_NTH_LOOP ))
	{
		next unless (defined $NEW_SET{$_});
		next if ($NEW_SET{$_} =~ /^\d+$/);
		
		# Permits simple expressions
		if ($NEW_SET{$_} =~ /^[\d\-+*\/\.\,]+$/)
		{
			my $new_value = eval $NEW_SET{$_};
			if ($new_value =~ /^(\d+)(\.\d+)?$/)
			{
				#print "$new_value = $1 (stripped $2);\n";
				$NEW_SET{$_} = $1;
			}
			else
			{
				&log("Error: Setting $_; evaluation of expression '$NEW_SET{$_}' failed $!", 1);				
				$errors++;
			}
			next;
		}
		
		&log("Error: Setting $_ requires a numerical integer value, has value '$NEW_SET{$_}'", 1);
		$errors++;
	}
	
	##
	## Checking >1 values
	##
	
	foreach (qw(REFRESH_RATE POLL_LINKINFO_EVERY_NTH_LOOP RRD_STEP
				GRAPH_WIDTH GRAPH_HEIGHT))
	{
		next unless (defined $NEW_SET{$_});
		next if ($NEW_SET{$_} >= 1);
		
		&log("Error: Setting $_ requires a numerical integer value larger that zero, has value '$NEW_SET{$_}'", 1);
		$errors++;
	}
	
	##
	## If SINGLE_LOOP is set then RETENTION_FILE must also be set
	## since the script won't loop but exit after first pass
	##
	
	if ($NEW_SET{'SINGLE_LOOP'})
	{
		unless ($NEW_SET{'RETENTION_FILE'})
		{
			# Previous data isn't preserved otherwise
			&log("Error: SINGLE_LOOP depends on RETENTION_FILE", 1);
			$errors++;
		}
		else
		{
			if ($NEW_SET{'RETENTION_EVERY_NTH_LOOP'} != 1)
			{
				# Previous data isn't preserved otherwise
				&log("SINGLE_LOOP; forcing RETENTION_EVERY_NTH_LOOP = 1", $NEW_SET{'VERBOSE'} >= 3);
				$NEW_SET{'RETENTION_EVERY_NTH_LOOP'} = 1;
			}
			
			# POLL_LINKINFO_EVERY_NTH_LOOP 
			foreach (qw(GRAPH_BPS_EVERY_NTH_LOOP GRAPH_PPS_EVERY_NTH_LOOP POLL_MACS_EVERY_NTH_LOOP))
			{
				if ($NEW_SET{$_} > 1)
				{
					# These must be performed every loop
					&log("SINGLE_LOOP; reducing $_ to 1", $NEW_SET{'VERBOSE'} >= 3);
					$NEW_SET{$_} = 1;
				}
			}
			
			if ($NEW_SET{'REFRESH_RATE'} > 30)
			{
				# If we must sleep the first loop, don't sleep too long
				&log("SINGLE_LOOP; reducing REFRESH_RATE to 30", $NEW_SET{'VERBOSE'} >= 3);
				$NEW_SET{'REFRESH_RATE'} = 30;
			}
			
		}
	}
	
	##
	## Add BASEDIR if given path isn't absolute
	##
	
	foreach (qw(RETENTION_FILE COOKIE_FILE HTML_FILE LOG_FILE WRAPS_FILE RRD_FILE_PREFIX))
	{
		next unless (defined $NEW_SET{$_});
		next if ($NEW_SET{$_} =~ /^\//);
		
		$NEW_SET{$_} = $SYSTEM->{'BASEDIR'} . $NEW_SET{$_};
	}
	
	##
	## Checking if RRDs should be written and the module can be loaded
	##
	
	if ($NEW_SET{'RRD_FILE_PREFIX'})
	{
		##
		## Shall log to RRD
		##
		
		eval('use RRDs;');
		if ($@)
		{
			##
			## Unable to load the RRDs-module!
			##
			
			$NEW_SET{'RRD_FILE_PREFIX'} = '';
			&log("Setting 'RRD_FILE_PREFIX' is set, but the perl-module RRDs.pm could not be loaded! Disabling RRD", 1);
		}
	}
	
	##
	## Shall some ports be excluded from RRD and graphing?
	##
	
	foreach my $portnum (split(/\s*,\s*/, $NEW_SET{'RRD_SKIP_THESE_PORTS'}))
	{
		# Rewrites the list for easy checking
		$NEW_SET{"RRD_SKIP_PORT_$portnum"} = 1;
	}
	
	##
	## If hiding inactive ports, always show these ports
	##
	
	foreach my $portnum (split(/\s*,\s*/, $NEW_SET{'ALWAYS_SHOW_THESE_PORTS'}))
	{
		# Rewrites the list for easy checking
		$NEW_SET{"ALWAYS_SHOW_PORT_$portnum"} = 1;
	}
	
	##
	## Parsing ports and graphs
	##
	
	my %PORTS = ();
	my %CHAINS = ();
	my @GRAPHS = ();
	my @RRD_RRA = ();
	my %MACS = ();
	
	foreach (sort keys %NEW_SET)
	{
		#print "$_\n";
		
		if (/^(PORT|TRUNK|INTERFACE)\s*(.+)$/)
		{
			##
			## This is a switch PORT/TRUNK or a local interface
			## PORT 1 = Router
			##
			
			my $port = $2;
			$port = "T$2" if ($1 eq 'TRUNK');
			#$port = "I$2" if ($1 eq 'INTERFACE');
			
			# Skips already defined ports
			if ($PORTS{$port})
			{
				&log("Warn: Port $port is already defined, ignoring this definition", 1);
				next;
			}
			
			# Memorizing this name
			$PORTS{$port} = $NEW_SET{$_};
		}
		elsif (/^CHAIN\s*(\d+)\s*(.+)$/)
		{
			##
			## This is a CHAIN
			## CHAIN 1 joxx:-1 = Router
			##
			
			my $port = "C$1";
			
			# Skips already defined ports
			if ($PORTS{$port})
			{
				&log("Warn: Port $port is already defined, ignoring this definition", 1);
				next;
			}
			
			$CHAINS{$port} = $2;
			
			# Memorizing this name
			$PORTS{$port} = $NEW_SET{$_};
		}
		elsif (/^GRAPH\s*(\d+)\s*\"(.+)\"$/)
		{
			##
			## This is a GRAPH
			## GRAPH "Past 3 hours"	= 60*60*3
			##
			
			my $index = $1;
			my $name = $2;
			
			# Permits simple expressions
			if ($NEW_SET{$_} =~ /^[\d\-+*\/\.\,]+$/)
			{
				my $new_value = eval $NEW_SET{$_};
				if ($new_value =~ /^(\d+)(\.\d+)?$/)
				{
					#print "$new_value = $1 (stripped $2);\n";
					$NEW_SET{$_} = $1;
				}
				else
				{
					&log("Error: Setting $_; evaluation of expression '$NEW_SET{$_}' failed $!", 1);				
					$errors++;
					next;
				}
			}
			
			# Memorizing this graph
			push (@GRAPHS, [$index, $name, $NEW_SET{$_}] );
		}
		elsif (/^RRD_RRA/)
		{
			##
			## This is a RRD RRA-statement
			## RRD_RRA 1 = RRA:AVERAGE:0.5:1:360
			##
			
			# Memorizing this RRA-definition
			push (@RRD_RRA, $NEW_SET{$_});
		}
		elsif (/^MAC\s+([a-fA-F0-9\.\:\- ]+)$/)
		{
			##
			## This is the description for a mac-address
			## MAC 00:11:22:33:44:55 = John's printer
			##
			
			my $mac = $1;
			$mac =~ s/\.\:\- //g;	# Keeping regular chars to allow different mac-syntaxes
			$mac =~ /^(..)(..)(..)(..)(..)(..)$/ || next;
			$mac = uc "$1:$2:$3:$4:$5:$6";
			
			# Memorizing this description
			$MACS{$mac} = $NEW_SET{$_};
		}
	}
	
	if ($NEW_SET{'SYNC_SLEEP_TO_RRD'} && ($NEW_SET{'RRD_STEP'} != $NEW_SET{'REFRESH_RATE'}))
	{
		&log("Can only sync sleep to RRD_STEP if REFRESH_RATE == RRD_STEP", 1);
		$NEW_SET{'SYNC_SLEEP_TO_RRD'} = 0;
	}
	
	# No ports defined?
	unless (%PORTS)
	{
		&log("Err:  No PORTs to poll is defined!", 1);
		$errors++;
	}
	
	# Memorizing the ports
	$NEW_SET{'PORTS'} = \%PORTS;
	
	# No graphs defined?
	if (($NEW_SET{'GRAPH_BPS_EVERY_NTH_LOOP'} || $NEW_SET{'GRAPH_PPS_EVERY_NTH_LOOP'}) && !@GRAPHS)
	{
		&log("Err:  No GRAPHs to create is defined!", 1);
		$errors++;
	}
	
	# Is selected GRAPH_OVERVIEW defined? FIXME
	#if ($NEW_SET{'GRAPH_RRD_EVERY_NTH_LOOP'} && !defined $GRAPHS[ $NEW_SET{'GRAPH_OVERVIEW'}-1 ])
	#{
	#	&log("Err:  Selected GRAPH_OVERVIEW $NEW_SET{'GRAPH_OVERVIEW'} does not exist!", 1);
	#	$errors++;
	#}
	
	# Memorizing the ports
	$NEW_SET{'GRAPHS'} = \@GRAPHS;

	# No RRD RRAs defined?
	if ($NEW_SET{'RRD_FILE_PREFIX'} && !@RRD_RRA)
	{
		&log("Err:  No RRD_RRAs is defined! Cannot create RRD-files without any", 1);
		$errors++;
	}
	
	# Memorizing the RRA definitions
	$NEW_SET{'RRD_RRA'} = \@RRD_RRA;
	
	# Memorizing the chains
	$NEW_SET{'CHAINS'} = \%CHAINS;
	
	# Memorizing the macs
	$NEW_SET{'MACS'} = \%MACS;
	
	##
	## Returns a 2 if basic errors are encountered
	##
	
	return 2 if ($errors);
	
	# Activating settings
	$SET = \%NEW_SET;
	close LOG;
	
	&show_settings() if ($SET->{'VERBOSE'} >= 4);
	
	# 1 = Settings changed
	return 1;
}

sub read_settings
{
	my $file = shift;
	my $NEW_SET = shift;
	
	&log("Reading settings from $file...", $SET->{'VERBOSE'} >= 2);
	
	my @INCLUDES = ();
	
	open (SET, $file) || return 0;
	while (<SET>)
	{
		next if (/^#/);		# Ignoring comments
		tr/\n\r//d;			# Stripping linefeeds
		s/\s+$//;			# Stripping trailing white spaces
		next unless ($_);	# Ignores empty lines
		
		if (/^INCLUDE\s+(.+?)\s*$/)
		{
			# Include settings from other file
			my $file = $1;
			$file = $SYSTEM->{'BASEDIR'} . $file unless ($file =~ /^\//);
			push (@INCLUDES, $file);
			next;
		}
		
		my ($name, $value) = split(/\s*=\s*/, $_, 2);
		$NEW_SET->{$name} = $value;
	}
	
	# Adding this modified-time to the total
	$NEW_SET->{'_modified'} += (stat(SET))[9];
	push (@{$SYSTEM->{'SETTINGS_FILES'}}, $file);
	#print "READ  $file - $NEW_SET->{'_modified'}\n";
	close SET;
	
	foreach (@INCLUDES)
	{
		&read_settings($_, $NEW_SET) || &log("Unable to INCLUDE settings-file '$_', ignoring it!", 1);
	}
	
	return 1;
}

##
## Show all settings
##

sub show_settings
{
	foreach (sort keys %$SET)
	{
		my $pad = ' ' x (30 - length($_));
		print "$_$pad = $SET->{$_}\n";
	}
}

##
## Log and possibly print
##

sub log
{
	my $text = shift;
	my $print_too = shift;
	
	print "$text\n" if ($print_too);
	
	unless (fileno(LOG))
	{
		return unless ($SET->{'LOG_FILE'});
		open (LOG, ">>$SET->{'LOG_FILE'}") || die("Unable to write to $SET->{'LOG_FILE'}!");
		
		select LOG;
		$| = 1;
		select STDOUT;
	}

	print LOG "$TIME{'date_time'}: $text\n";
}

##
## Log and then die
##

sub log_n_die
{
	my $text = shift;
	&log($text);
	die $text;
}

##
## Create the HTML-page
##

sub generate_html
{
	#print "HTML refresh: $refresh\n";
	
	open (HTML, ">$SET->{'HTML_FILE'}") || warn "Unable to write to html-file '$SET->{'HTML_FILE'}'!";
	print HTML <<END;
<html>
<head>
<title>Statistics for $SET->{'SWITCH'} - $SYSTEM->{'sysname'} $SYSTEM->{'location'}</title>
<meta http-equiv="Refresh" content="$SET->{'REFRESH_RATE'}">
</head>

<body bgcolor="White" text="Black" link="Black" vlink="Black">

<center>
<p>
<table border="0" cellpadding="4" cellspacing="0" style="font-family: Verdana, Geneva, Arial, Helvetica, sans-serif; font-size: 10;" width="980">

<tr align=left valign=bottom>
<th align=right rowspan=2 bgcolor="silver">#</th>
<th rowspan=2 colspan=2 bgcolor="silver">Equipment</th>
<th colspan=6 bgcolor="lime"><a title="When polling a local interface, it's traffic transmitted from it">OUT: Switch -&gt; Equipment:</a></th>
<th colspan=2 bgcolor="red">Errors</th>
<td bgcolor="silver"></td>
<th colspan=6 bgcolor="lime"><a title="When polling a local interface, it's traffic recieved by it">IN: Equipment -&gt; Switch:</a></th>
<th colspan=5 bgcolor="red">Errors</th>
</tr>
<tr align=left valign=bottom>
<th bgcolor="lime" colspan=2><a title="goodOctetSND*8 (bits/s), txGood (packets/s)">Rate</a></th>
<th bgcolor="lime" colspan=2><a title="goodOctetSND (bytes), txGood (packets)">TOT</a></th>
<th bgcolor="lime"><a title="txBroadcast (packets)">Bcast</a></th>
<th bgcolor="lime"><a title="txMulticast (packets)">Mcast</a></th>
<th bgcolor="red"><a title="txDrop (packets)">Drop</a></th>
<th bgcolor="red"><a title="txCollisions (packets)">Coll</a></th>
<td bgcolor="silver"></td>

<th bgcolor="lime" colspan=2><a title="goodOctetRCV*8 (bits/s), rxGood (packets/s)">Rate</a></th>
<th bgcolor="lime" colspan=2><a title="goodOctetRCV (bytes), rxGood (packets)">TOT</a></th>
<th bgcolor="lime"><a title="rxBroadcast (packets)">Bcast</a></th>
<th bgcolor="lime"><a title="rxMulticast (packets)">Mcast</a></th>
<th bgcolor="red"><a title="rxFCS (packets)">FCS</a></th>
<th bgcolor="red"><a title="rxUnderSize (packets)">Usize</a></th>
<th bgcolor="red"><a title="rxOverSize (packets)">Osize</a></th>
<th bgcolor="red"><a title="rxFragment (packets)">Frag</a></th>
<th bgcolor="red"><a title="rxJabber (packets)">Jabb</a></th>
</tr>
END
	
	foreach my $portnum (@do_ports)
	{
		##
		## Each port
		##
		
		my $STAT = $STATS{$portnum};
		my $OLD_STAT = $OLD_STATS{$portnum};
		my $PORT = $PORTS{$portnum};
		my $is_err = 0;
		
		# Hides ports that are not active
		next if ($SET->{'HIDE_INACTIVE_PORTS'} && !$PORT->{'is_active'} && !$SET->{"ALWAYS_SHOW_PORT_$portnum"});
		
		##
		## Shall we add a link and/or title?
		##
		
		my $portnum_link = '';
		my $portnum_title = '';
		
		if ($PORT->{'rrd_bps_file'})
		{
			# Link to BPS-graph
			$portnum_link = $PORT->{'rrd_bps_file'};
			$portnum_title = "Click to view graphs";
		}
		elsif ($PORT->{'rrd_pps_file'})
		{
			# Link to PPS-graph
			$portnum_link = $PORT->{'rrd_pps_file'};
			$portnum_title = "Click to view graphs";
		}
		
		if ($portnum =~ /^C\d+$/)
		{
			# Chain
			my @def = split(/\s*,\s*/, ${$SET->{'CHAINS'}}{$portnum});
			
			$portnum_title .= ", " if ($portnum_title);
			$portnum_title .= "OUT = chain $def[0] row $def[1]";
			$portnum_title .= ", IN = chain $def[2] row $def[3]" if ($def[2]);
		}
		
		$portnum_link  = "href=\"$portnum_link\"" if ($portnum_link);
		$portnum_title = "title=\"$portnum_title\"" if ($portnum_title);
		
		my $portdesc = $do_ports{$portnum};
		
		my $macs = '';
		if ($PORT->{'macs'})
		{
			if ($SET->{'POLL_MACS_EVERY_NTH_LOOP'})
			{
				$macs = "&nbsp;<a style=\"color: silver;\" title=\"$PORT->{'macs'}\">($PORT->{'macs_count'} mac)</a>";
				$macs =~ s/\n/, \n/g;
			}
			else
			{
				# Helping with cleanup...
				delete $PORT->{'macs'};
				delete $PORT->{'macs_count'};
			}
		}
		
		##
		## Writing
		##
		
		print HTML <<END;
<tr align=right valign=top onmouseover="this.style.backgroundColor='$TR_COLOR'" onmouseout="this.style.backgroundColor=''">
<th bgcolor="silver"><a $portnum_link $portnum_title>$portnum</a></th>
<td align=left>$portdesc</td>
<td align=right>$macs</td>
END
			
		foreach (qw(
delta_goodOctetSND delta_txGood goodOctetSND txGood txBroadcast txMulticast
is_err txDrop txCollisions no_err
space
delta_goodOctetRCV delta_rxGood goodOctetRCV rxGood rxBroadcast rxMulticast
is_err rxFCS rxUnderSize rxOverSize rxFragment rxJabber no_err
			))	# txError rxError 
		{
			##
			## Visar vrden
			##
			
			if ($_ eq 'space')
			{
				print HTML "<td></td>";
				next;
			}
			elsif ($_ eq 'is_err')
			{
				$is_err = 1;
				next;
			}
			elsif ($_ eq 'no_err')
			{
				$is_err = 0;
				next;
			}
			
			my $value = my $real_value = $STAT->{$_};
			my $diff = $value - $OLD_STAT->{$_};
			my $make_boldred = ($is_err && defined $OLD_STAT->{$_} && ($diff > 0));
			$diff = ($make_boldred) ? " (+$diff)" : '';
			
			if (/^delta_/)
			{
				# Delta-variable, add prefix and do some extra work
				if ($value > 0)
				{
					if (/Octet/)
					{
						# Bytes, shall be bits
						$value *= 8;	# We want bit/s not byte/s
						$real_value = $value;
						
						$value = &shorten_num($value, undef, 1024) . 'bps';
						
						$value = "<font color=\"silver\">$value</font>" if ($SET->{'GREY_BELOW_BPS'} && $real_value < $SET->{'GREY_BELOW_BPS'});
					}
					else
					{
						# Packets
						$value = &shorten_num($value, undef, 1000) . 'pps';
						
						$value = "<font color=\"silver\">$value</font>" if ($SET->{'GREY_BELOW_PPS'} && $real_value < $SET->{'GREY_BELOW_PPS'});
					}
				}
				else
				{
					$value = '';
				}
			}
			elsif (/^(goodOctetRCV|rxGood|txGood|goodOctetSND)$/)
			{
				# Normal variable, but we shall add prefix for clarity
				if ($value > 0)
				{
					if (/Octet/)
					{
						# Bytes
						$value = &shorten_num($value, undef, 1024) . 'B';
					}
					else
					{
						# Packets
						$value = &shorten_num($value, undef, 1000) . 'p';
					}
				}
				else
				{
					$value = "<font color=\"silver\">$value</font>";
				}
			}
			else
			{
				# Other variables, shorten numbers, no prefix
				# These are all packets, so we use 1000 as base
				if ($value > 0)
				{
					$value = &shorten_num($value,4,1000);
				}
				else
				{
					$value = "<font color=\"silver\">$value</font>";
				}
			}
			
			# If a error-counter increases it will be shown in bold red
			$value = "<b><font color=red>$value</font></b>" if ($make_boldred);
			
			$value = "<b>$value</b>" if ($SET->{'BOLD_ABOVE_BPS'} && $real_value > $SET->{'BOLD_ABOVE_BPS'} && $value =~ /bps$/);
			$value = "<b>$value</b>" if ($SET->{'BOLD_ABOVE_PPS'} && $real_value > $SET->{'BOLD_ABOVE_PPS'} && $value =~ /pps$/);
			
			# Exact value for mouse-over
			my $value_sep = &add_sep($real_value);
			
			# Extra text to show?
			my $info = ($PORT->{"$_:INFO"}) ? "   \n".$PORT->{"$_:INFO"} : '';
			
			# Writing da stuff
			print HTML "<td><a title=\"$value_sep$diff$info\">$value</a></td>";
		}
		
		print HTML "</tr>\n";
	}
	
	print HTML <<END;
</table>

<br><br><br>
<table border="0" cellpadding="4" cellspacing="0" style="font-family: Verdana, Geneva, Arial, Helvetica, sans-serif; font-size: 10;" width=980>

<tr align=left valign=bottom>
<th align=right rowspan=2 bgcolor="silver">#</th>
<th rowspan=2 bgcolor="silver">Equipment</th>
<th bgcolor="lime"></th>
<th bgcolor="silver" width=0></th>
<th colspan=23 bgcolor="lime">IN: Equipment -> Switch: Amount of packets per framesize</th>
</tr>
<tr align=center valign=bottom>
<th bgcolor="lime">Link</th>
<th bgcolor="silver" width=0></th>
<th bgcolor="lime" colspan=3><a title="rx64Octets (packets, percent, packets/s)">64 byte</a></th>
<td bgcolor="silver" width=0></td>
<th bgcolor="lime" colspan=3><a title="rx65TO127Octets (packets, percent, packets/s)">65-127 byte</a></th>
<td bgcolor="silver" width=0></td>
<th bgcolor="lime" colspan=3><a title="rx128TO255Octets (packets, percent, packets/s)">128-255 byte</a></th>
<td bgcolor="silver" width=0></td>
<th bgcolor="lime" colspan=3><a title="rx256TO511Octets (packets, percent, packets/s)">256-511 byte</a></th>
<td bgcolor="silver" width=0></td>
<th bgcolor="lime" colspan=3><a title="rx512TO1023Octets (packets, percent, packets/s)">512-1024 byte</a></th>
<td bgcolor="silver" width=0></td>
<th bgcolor="lime" colspan=3><a title="rx1024ToMa (packets, percent, packets/s)">1025- byte</a></th>
</tr>
END
	
	foreach my $portnum (@do_ports)
	{
		##
		## Each port
		##
		
		next unless ($portnum =~ /^T?\d+$/);	# No such info for local interfaces
		
		my $STAT = $STATS{$portnum};
		my $OLD_STAT = $OLD_STATS{$portnum};
		my $PORT = $PORTS{$portnum};
		
		# Hides ports that are not up, if wanted
		next if ($SET->{'HIDE_INACTIVE_PORTS'} && !$PORT->{'is_active'} && !$SET->{"ALWAYS_SHOW_PORT_$portnum"});
		
		##
		## Shall we add a link to graphs?
		##
		
		my $graph_link = '';
		
		if ($PORT->{'rrd_bps_file'})
		{
			# Link to BPS-graph
			$graph_link = $PORT->{'rrd_bps_file'};
		}
		elsif ($PORT->{'rrd_pps_file'})
		{
			# Link to PPS-graph
			$graph_link = $PORT->{'rrd_pps_file'};
		}
		
		$graph_link = "href=\"$graph_link\" title=\"Click to view graphs\"" if ($graph_link);
		
		##
		## Make linkinfo bold if it recently changed
		##
		
		my $link = $PORT->{'link'};
		$link = 'up?' if ($link eq 'down' && $PORT->{'is_active'});
		$link = "<b>$link</b>" if ($STAT->{'link_changed'});
		
		##
		## Writing
		##
		
		print HTML <<END;
<tr align=right valign=top onmouseover="this.style.backgroundColor='$TR_COLOR'" onmouseout="this.style.backgroundColor=''">
<th bgcolor="silver"><a $graph_link>$portnum</a></th>
<td align=left>$do_ports{$portnum}</td>
<td align=right><a title="$PORT->{'link_change_at'}">$link</a></td>
<td></td>
END
		
		my $tot_packets = 0;
		foreach (qw( rx64Octets rx65TO127Octets rx128TO255Octets rx256TO511Octets rx512TO1023Octets rx1024ToMa ))
		{
			$tot_packets += $STAT->{$_};
		}
		
		foreach (qw( rx64Octets rx65TO127Octets rx128TO255Octets rx256TO511Octets rx512TO1023Octets rx1024ToMa ))
		{
			##
			## The amount of packets
			##
			
			my $tmp = $STAT->{$_};
			my $value_sep = &add_sep($tmp);
			my $value = &shorten_num($tmp, 4, 1000);
			my $proc = 0;
			eval { $proc = int($tmp / $tot_packets * 100) };
			
			if ($tmp eq '0' || ( $SET->{'GREY_BELOW_PROC'} && ($proc < $SET->{'GREY_BELOW_PROC'})))
			{
				$value = "<font color=\"silver\">$value</font>";
			}
			
			print HTML "<td><a title=\"$value_sep\">$value</a></td>";
			
			##
			## Percent of total
			##
			
			$proc = (($proc >= $SET->{'GREY_BELOW_PROC'}) || !$SET->{'GREY_BELOW_PROC'}) ? "$proc\%" : "<font color=\"silver\">$proc%</font>";
			
			print HTML "<td>$proc</td>";
			
			##
			## Delta
			##
			
			my $delta = $STAT->{"delta_$_"};
			
			if ($delta > 0)
			{
				$value = &shorten_num($delta, undef, 1000) . 'pps';
				$value = "<b>$value</b>" if ($SET->{'BOLD_ABOVE_PPS'} && $delta > $SET->{'BOLD_ABOVE_PPS'} && $value =~ /pps$/);
				$value = "<font color=\"silver\">$value</font>" if ($SET->{'GREY_BELOW_PPS'} && $delta < $SET->{'GREY_BELOW_PPS'});
			}
			else
			{
				$value = '';
			}
			
			$value_sep = &add_sep($delta);
			
			print HTML "<td><a title=\"$value_sep\">$value</a></td>";
			print HTML "<td></td>" unless ($_ eq 'rx1024ToMa');
		}
		
		print HTML "</tr>\n";
	}
	
	my $uptime = '';
	if ($SYSTEM->{'uptime'} > 0)
	{
		$uptime = ", uptime ".&pretty_time($SYSTEM->{'uptime'});
	}
	
	# Refresh rate
	my $refresh_rate = int $SYSTEM->{'real_refresh_rate'} || $SET->{'REFRESH_RATE'};
	
	my $created = localtime;
	my $tmp = $looptime;
	$tmp =~ s/(\.\d\d).+$/$1/;
	
	# ServiceTag $SYSTEM->{'servicetag'}, Serial $SYSTEM->{'serial'}
	print HTML <<END;
</table>

<p>
<div style="font-family: Verdana, Geneva, Arial, Helvetica, sans-serif; font-size: 9;">
$created - loop $loop - The rates are averages since last loop, around $refresh_rate seconds (loop took ${tmp}s)<br>
<a href="http://$SET->{'SWITCH'}" target="delladm" title="Asset $SYSTEM->{'assettag'}">$SYSTEM->{'model'}</a>, name $SYSTEM->{'sysname'}, location $SYSTEM->{'location'}, mac $SYSTEM->{'macaddress'}, fw $SYSTEM->{'firmware'}$uptime
<p>
<font color=silver>Generated by $progname $progver written by Joakim Andersson - http://www.iqcc.se<br>
Using switch-specific module $modname $modver written by $modcreator</font>
</div>

</center>

</body>
</html>
END
	
	close HTML;
}

##
## Get system and portinfo from switch
##

sub get_system_info
{
	print "Polling link- and system info...\n" if ($SET->{'VERBOSE'} >= 2);
		
	my $ports = &switch_system_info() || return 0;
	
	foreach my $portnum (sort { $ports->{$a}->{'order'} <=> $ports->{$b}->{'order'} || $a <=> $b } keys %$ports)
	{
		my $PORT = $PORTS{$portnum} ||= { 'link_changes' => 0, 'link_change_at' => 'No change detected' };
		
		my $link = $ports->{$portnum}->{'link'};
		my $is_active = $ports->{$portnum}->{'is_active'};
		
		if ($PORT->{'link'} && ($link ne $PORT->{'link'}))
		{
			&log("$do_ports{$portnum} (port $portnum): Change of link detected; $PORT->{'link'} -> $link", $SET->{'VERBOSE'} >= 3);
			
			# To show in the HTML
			$PORT->{'link_changes'}++;
			$PORT->{'link_change_at'} = "Change detected $TIME{'date_time'}; \n$PORT->{'link'} -> $link. $PORT->{'link_changes'} changes";
			
			# Triggers the link to be shown in bold
			$STATS{$portnum}->{'link_changed'} = 1;
		}
		
		$PORT->{'link'} = $link;
		$PORT->{'is_active'} = $PORT->{'has_link'} = $is_active;
		
		if ($SET->{'AUTOADD_ACTIVE_PORTS'} && $is_active && !defined $SET->{'PORTS'}->{$portnum})
		{
			##
			## We shall autoadd active ports
			##
			
			if ($portnum =~ /^T(\d+)$/)
			{
				$SET->{'PORTS'}->{$portnum} = "LAG $1";
				&log("Auto-added trunk $1 as $portnum, link detected", $SET->{'VERBOSE'} >= 1);
			}
			else
			{
				$SET->{'PORTS'}->{$portnum} = "Port $portnum";
				&log("Auto-added port $portnum, link detected", $SET->{'VERBOSE'} >= 1);
			}
		}
	}
	
	##
	## Speeds up things
	##
	
	%do_ports = %{$SET->{'PORTS'}};	# Ports on the switch
	@do_ports = sort { length($a) <=> length($b) || $a cmp $b } keys %do_ports;
	
	##
	## Find longest portname
	##
	
	$SYSTEM->{'maxlen_portname'} = 10;
	foreach my $portnum (@do_ports)
	{
		my $this_len = length($do_ports{$portnum});
		$SYSTEM->{'maxlen_portname'} = $this_len if ($SYSTEM->{'maxlen_portname'} < $this_len);
	}
	
	return 1;
}

##
## Login into a switch
##

sub login
{
	print "Logging into the switch...\n" if ($SET->{'VERBOSE'} >= 2);
	$SYSTEM->{'cookie'} = &switch_login();
	
	unless ($SYSTEM->{'cookie'})
	{
		&log_n_die("Unable to login! Did not recieve cookie");
	}
	
	&log("Logged in, using session-cookie: $SYSTEM->{'cookie'}", $SET->{'VERBOSE'});
	
	##
	## Caching the session for future runs
	## Prevents hanging of the web-interface if script is restarted un too often
	##
	
	if ($SET->{'COOKIE_FILE'})
	{
		open (COOKIE, ">$SET->{'COOKIE_FILE'}") || &log_n_die("Unable to save the cookie to file $SET->{'COOKIE_FILE'}!");
		print COOKIE $SYSTEM->{'cookie'};
		close COOKIE;
	}
	
	#if ($SET->{'RETENTION_FILE'})
	#{
	#	&save_retention();
	#}
}

##
## Log out from a switch
##

sub logout
{
	&switch_logout();
	
	if ($SYSTEM->{'cookie'})
	{
		&log("Unable to logout, cookie is still set. Setting ALWAYS_LOGOUT = 0", 1);
		$SET->{'ALWAYS_LOGOUT'} = 0;
	}
	elsif ($SET->{'COOKIE_FILE'})
	{
		open (COOKIE, ">$SET->{'COOKIE_FILE'}") || &log_n_die("Unable to save the cookie to file $SET->{'COOKIE_FILE'}!");
		print COOKIE $SYSTEM->{'cookie'};
		close COOKIE;
	}
	
	#if ($SET->{'RETENTION_FILE'})
	#{
	#	&save_retention();
	#}
}

##
## Load retention
##

sub load_retention
{
	return 0 unless ($SET->{'RETENTION_FILE'});
	
	my $h = '';
	$h = lock_retrieve($SET->{'RETENTION_FILE'}) if (-e $SET->{'RETENTION_FILE'});
	
	unless ($h)
	{
		my $bak = $SET->{'RETENTION_FILE'}.'.bak';
		if (-e $bak)
		{
			# A backup exist
			&log("Unable to load RETENTION_FILE '$SET->{'RETENTION_FILE'}', trying backup", 1);
			
			$h = lock_retrieve($SET->{'RETENTION_FILE'}.'.bak');
			unless ($h)
			{
				# Nope!
				&log("Unable to load backup RETENTION_FILE '$bak'!", 1);
				return 0;
			}
		}
		else
		{
			# No backup exist, fail silently
			&log("Unable to load RETENTION_FILE '$SET->{'RETENTION_FILE'}', first run?", 1);
			return 0;
		}
	}
	
	#use Data::Dumper;
	#print Dumper($h);
	
	my @keys = keys %$h;	# (@keys)
	&log("Loaded retention", 1) if (@keys);
	
	%STATS		= %{$h->{'STATS'}}		if (ref $h->{'STATS'} eq 'HASH');
	%OLD_STATS	= %{$h->{'OLD_STATS'}}	if (ref $h->{'OLD_STATS'} eq 'HASH');
	%PORTS		= %{$h->{'PORTS'}}		if (ref $h->{'PORTS'} eq 'HASH');
	%WRAPS		= %{$h->{'WRAPS'}}		if (ref $h->{'WRAPS'} eq 'HASH');
	$SYSTEM->{'cookie'} = $h->{'cookie'} if ($h->{'cookie'});
	
	return 1;
}

##
## Save retention
##

sub save_retention
{
	return 0 unless ($SET->{'RETENTION_FILE'});
	
	my %hash = (
		'STATS'		=> \%STATS,		# To maintain delta-stats 
		'OLD_STATS' => \%OLD_STATS,	# To cache previous values
		'PORTS'		=> \%PORTS,		# To maintain non-statistics info about each port
		'WRAPS'		=> \%WRAPS,		# To maintain counter-wraps
		'cookie'	=> $SYSTEM->{'cookie'}
	);
	
	# Making backup. UGLY, but will do, hopefully
	`cp "$SET->{'RETENTION_FILE'}" "$SET->{'RETENTION_FILE'}.bak" >/dev/null`;
	
	# Saving
	if (lock_store(\%hash, $SET->{'RETENTION_FILE'}))
	{
		my $size = -s $SET->{'RETENTION_FILE'};
		&log("Saved retention ($size bytes)", $SET->{'VERBOSE'} >= 2);
		return 1;
	}
	
	# Failed!
	&log("Unable to save retention to RETENTION_FILE '$SET->{'RETENTION_FILE'}'! $@", 1);
	return 0;
}

1;
